#
# Coupled tanks example: Direct collocation
# Master thesis attachment
# Konstantin Nalum 2013
#

import os
import sys
import numpy
from numpy import *
from matplotlib.pyplot import *
import zipfile
import time
import scipy.io

os.system("CLS")
from pymodelica import compile_jmu,compile_fmu,compile_fmux
from pyjmi import CasadiModel, JMUModel
from casadi import *
from pyfmi import FMUModel
from pymodelica import compile_fmu
from pyfmi import FMUModel


tfinal = 600


fmux = compile_fmux("ocp", "tank.mop")				# Compile model
tank = CasadiModel(fmux, scale_variables=True)		# Import to CasADi and scale model

opt_opts = tank.optimize_options()					# Set options for optimization

N = 50							
K = 3
Nval_points = 20

opt_opts['n_e'] = N
opt_opts['n_cp'] = K
opt_opts['result_mode']='element_interpolation'		# Obtain interpolated variable profiles from optimization
opt_opts['n_eval_points'] = Nval_points 			# Number of interpolated points (smoothness of interpolated profiles)


res = tank.optimize(algorithm="LocalDAECollocationAlg",options=opt_opts)

# Obtain results from optimization
z1_o=res['z1']
z2_o=res['z2']
u_o = res['u']
t_o=res['time']


u_traj = transpose(vstack((t_o,u_o))) 		# Create u profile for inputting to simulation
input_object = (['u'], u_traj)

fmu_tank = compile_fmu("tank","tank.mop")	# Compile model for simulation purpose
tank_sim = FMUModel(fmu_tank)

opts = tank_sim.simulate_options()
opts['solver'] = 'CVode'
opts['ncp'] = Nval_points*N-1
res = tank_sim.simulate(final_time=tfinal, input=input_object,options=opts)


# Plot optimized and simulated profiles
z1_s=res['z1']
z2_s=res['z2']
u_s = res['u']
t_s=res['time']

clf()
subplot(311)
plot(t_o,z1_o,'r--')
plot(t_s,z1_s,'r')

subplot(312)
plot(t_o,z2_o,'g--')
plot(t_s,z2_s,'g')

subplot(313)
plot(t_o,u_o,'b')
plot(t_s,u_s,'r--')

# Export data to MATLAB
# fname='C:/'
# scipy.io.savemat(fname, \
			# dict(	t_s=t_s,t_o=t_o,\
					# z1_s=z1_s,z2_s=z2_s,u_s=u_s,\
					# z1_o=z1_o,z2_o=z2_o,u_o=u_o))


# fname='C:/'
# scipy.io.savemat(fname, \
			# dict(	t_s=t_s,t_o=t_o,\
					# z1_s=z1_s,z2_s=z2_s,u_s=u_s,\
					# z1_o=z1_o,z2_o=z2_o,u_o=u_o))
					
show()	# Show plots