#
# Well-Flowline: Direct collocation
# Master thesis attachment
# Konstantin Nalum 2013
#

import os
import sys
import numpy
from numpy import *
from matplotlib.pyplot import *
import zipfile
import time
import scipy.io

os.system("CLS")
from pymodelica import compile_jmu,compile_fmu,compile_fmux
from pyjmi import CasadiModel, JMUModel
from casadi import *
from pyfmi import FMUModel

# Parameters

tfinal=6*3600			# Final time (needs to be changed inside MOP-file too)
u_min = 1
u_max = 2
u_tot_max = 7
q_o =  1
q_gi = 0.1

N = 20				# Number of elements
K = 3				# Number of collocation points in each element
IPOPT_TOL = 1e-1		# IPOPT tolerance
IPOPT_MAX_ITER = 1000	# IPOPT max iterations


# Finding initial guess by simulation

init_sim_fmu = compile_fmu("MODEL.well_flowline_init_optimization","modell.mop")
init_sim_model = FMUModel(init_sim_fmu)
init_res = init_sim_model.simulate(start_time=0, final_time=tfinal)



fmux = compile_fmux("MODEL.ocp", "modell.mop")				# Compile model
wf_system = CasadiModel(fmux, scale_variables=True)			# Import to CasADi and scale model

wf_system.set('u_min',u_min)
wf_system.set('u_max',u_max)
wf_system.set('q_o',q_o)
wf_system.set('q_gi',q_gi)
wf_system.set('u_tot_max',u_tot_max)

opt_opts = wf_system.optimize_options()						# Set options for optimization
opt_opts['n_e'] = N 
opt_opts['n_cp'] = K
opt_opts['init_traj'] = init_res.result_data
opt_opts['IPOPT_options']['max_iter'] = IPOPT_MAX_ITER
opt_opts['IPOPT_options']['tol'] = IPOPT_TOL


n_eval_points = 20
opt_opts['result_mode']="element_interpolation"				# Obtain interpolated variable profiles from optimization
opt_opts['n_eval_points'] = n_eval_points					# Number of interpolated points (smoothness of interpolated profiles)

# Input blocking feature (OPTIONAL)
# blk  = array(ones(N)*1) 
# blk = blk.astype(int)
# opt_opts['blocking_factors'] = blk

res = wf_system.optimize(algorithm="LocalDAECollocationAlg",options=opt_opts)



# Obtain results from optimization

w_gout1=res['wf.f1.w_gout']
w_oout1=res['wf.f1.w_oout']
w_wout1=res['wf.f1.w_wout']

w_gout2=res['wf.f2.w_gout']
w_oout2=res['wf.f2.w_oout']
w_wout2=res['wf.f2.w_wout']

u_opt1 = res['wf.w1.u_gi']
u_opt2 = res['wf.w2.u_gi']
u_opt3 = res['wf.w3.u_gi']
u_opt4 = res['wf.w4.u_gi']
u_opt5 = res['wf.w5.u_gi']

t_opt=res['time']

# Export data to MATLAB
# scipy.io.savemat('C:/', \
			# dict(	t_c_opt=t_opt,\
					# w_gout1_c_opt=w_gout1,w_oout1_c_opt=w_oout1,w_wout1_c_opt=w_wout1,\
					# w_gout2_c_opt=w_gout2,w_oout2_c_opt=w_oout2,w_wout2_c_opt=w_wout2,\
					# u_gi1_c_opt=u_opt1,u_gi2_c_opt=u_opt2,u_gi3_c_opt=u_opt3,u_gi4_c_opt=u_opt4,u_gi5_c_opt=u_opt5))

# Simulate real model with optimized control input

from matplotlib.pyplot import *
from pymodelica import compile_fmu
from pyfmi import FMUModel

import matplotlib.pyplot as plt
from numpy import *


fmu_name = compile_fmu("MODEL.full","modell.mop")	# Compile model for simulation purpose

vdp = FMUModel(fmu_name)

opts = vdp.simulate_options()
opts['solver'] = 'CVode'
opts['ncp'] = n_eval_points*N-1

# Create u profile for inputting to simulation
u_traj = transpose(vstack((t_opt,u_opt1,u_opt2,u_opt3,u_opt4,u_opt5)))

input_object = (['u1','u2','u3','u4','u5'], u_traj)
res = vdp.simulate(final_time=tfinal,input=input_object,options=opts)

w_gout1 = res['f1.w_gout']
w_oout1 = res['f1.w_oout']
w_wout1 = res['f1.w_wout']

w_gout2 = res['f2.w_gout']
w_oout2 = res['f2.w_oout']
w_wout2 = res['f2.w_wout']

u1 = res['u1']
u2 = res['u2']
u3 = res['u3']
u4 = res['u4']
u5 = res['u5']


t = res['time']

# Export data to MATLAB
# scipy.io.savemat('C:/', \
			# dict(	t_c=t,\
					# w_gout1_c=w_gout1,w_oout1_c=w_oout1,w_wout1_c=w_wout1,\
					# w_gout2_c=w_gout2,w_oout2_c=w_oout2,w_wout2_c=w_wout2,\
					# u_gi1_c=u1,u_gi2_c=u2,u_gi3_c=u3,u_gi4_c=u4,u_gi5_c=u5))



				

# Plot optimized and simulated profiles
clf()

subplot(311)
plot(t,w_gout1,'g')
plot(t,w_oout1,'k')
plot(t,w_wout1,'b')


subplot(312)
plot(t,w_gout2,'g')
plot(t,w_oout2,'k')
plot(t,w_wout2,'b')

subplot(313)
plot(t,u1,'r')
plot(t,u2,'g')
plot(t,u3,'b')
plot(t,u4,'m')
plot(t,u5,'k')


show()	# Show plots