#
# Well-Flowline: Multiple shooting
# Master thesis attachment
# Konstantin Nalum 2013
#

import os
import sys
import numpy as NP
from numpy import *
import matplotlib.pyplot as plt
import zipfile
import time

os.system("CLS")
from pymodelica import compile_jmu
from pyjmi import JMUModel
from casadi import *
from matplotlib.pyplot import *
from pymodelica import compile_fmu
from pyfmi import FMUModel
import numpy as np
import scipy.io

# Compile model and generate XML equations
curr_dir = os.path.dirname(os.path.abspath(__file__))
jmu_name = compile_jmu("MODEL.ocp", curr_dir+"/modell.mop",'optimica','ipopt',{'generate_xml_equations':True, 'generate_fmi_me_xml':False})
sfile = zipfile.ZipFile(curr_dir+'/MODEL_ocp.jmu','r')
mfile = sfile.extract('modelDescription.xml','.')

# Obtain Symbolic OCP representation
ocp = SymbolicOCP()
ocp.parseFMI('modelDescription.xml',{'scale_variables':True,'sort_equations':False})
ocp.sortType()


tf=6*3600		# Final time
nk=60			# Number of segments
nWells=5		# Number of wells
nFlowlines=2	# Number of flowlines

u_min 		= 	ocp.variable('u_min').getStart()		# Get bound from Optimica
u_max 		= 	ocp.variable('u_max').getStart()		# Get bound from Optimica
u_tot_max 	=	ocp.variable('u_tot_max').getStart()	# Get capacity from Optimica

#uMinCapacity <= Sum(u_gi) <= uMaxCapacity : For different intervals
uMinCapacity = 1*ones(nk)
uMaxCapacity = concatenate(((7)*ones(nk/2),\
							(7)*ones(nk/2)))		

IPOPT_TOL = 1e-1					# IPOPT tolerance (tol)
IPOPT_MAX_ITER = 150				# IPOPT maximum iterations

nStates=nWells*3+nFlowlines*6+1		# Model states + cost state


def toArray(v, der=False):			# Convert symbolic OCP variables to another format
  ret = []
  for i in v:
    if der:
      ret.append(i.der())
    else:
      ret.append(i.var())
  return array(ret,dtype=SX)
  

ocp.eliminateDependent()
ocp.makeExplicit()

z = toArray(ocp.xd)		# States
u = toArray(ocp.u)		# Control inputs
f=ocp.ode				# ODE RHS (model equations)

# Generate ode: f=(z,u)
ode=SXFunction(daeIn(z,[],u,[],[]),daeOut(f))


f_d = CVodesIntegrator(ode)						# Instantiate a CVodesIntegrator
f_d.setOption("abstol",1e-5) 					# Absolute tolerance
f_d.setOption("reltol",1e-5) 					# Relative tolerance	
f_d.setOption("steps_per_checkpoint",1000)		# Steps per checkpoint
f_d.setOption("tf",tf/nk) 						# Integration length for each u^l
f_d.init()



nv = nWells*nk + nStates*(nk)

V = msym("V", nv)								# Decision variable vector [u,z]

# Get the expressions for local variables
U = V[0:nk*nWells]
Z0 = V[nk*nWells+0:nv:nStates]

Z1 = V[nk*nWells+1:nv:nStates]
Z2 = V[nk*nWells+2:nv:nStates]
Z3 = V[nk*nWells+3:nv:nStates]

Z4 = V[nk*nWells+4:nv:nStates]
Z5 = V[nk*nWells+5:nv:nStates]
Z6 = V[nk*nWells+6:nv:nStates]

Z7 = V[nk*nWells+7:nv:nStates]
Z8 = V[nk*nWells+8:nv:nStates]
Z9 = V[nk*nWells+9:nv:nStates]

Z10 = V[nk*nWells+10:nv:nStates]
Z11 = V[nk*nWells+11:nv:nStates]
Z12 = V[nk*nWells+12:nv:nStates]

Z13 = V[nk*nWells+13:nv:nStates]
Z14 = V[nk*nWells+14:nv:nStates]
Z15 = V[nk*nWells+15:nv:nStates]

Z16 = V[nk*nWells+16:nv:nStates]
Z17 = V[nk*nWells+17:nv:nStates]
Z18 = V[nk*nWells+18:nv:nStates]
Z19 = V[nk*nWells+19:nv:nStates]
Z20 = V[nk*nWells+20:nv:nStates]
Z21 = V[nk*nWells+21:nv:nStates]

Z22 = V[nk*nWells+22:nv:nStates]
Z23 = V[nk*nWells+23:nv:nStates]
Z24 = V[nk*nWells+24:nv:nStates]
Z25 = V[nk*nWells+25:nv:nStates]
Z26 = V[nk*nWells+26:nv:nStates]
Z27 = V[nk*nWells+27:nv:nStates]




# Variable bounds initialized to +/- inf

VMIN = (-inf)*NP.ones(nv)
VMAX = (inf)*NP.ones(nv)

# Control bounds
VMIN[0:nk*nWells] = u_min
VMAX[0:nk*nWells] = u_max


# Initial conditions
L0=0

z10=ocp.variable('wf.w1.m_ga').getStart()
z20=ocp.variable('wf.w1.m_gt').getStart()
z30=ocp.variable('wf.w1.m_lt').getStart()

z40=ocp.variable('wf.w2.m_ga').getStart()
z50=ocp.variable('wf.w2.m_gt').getStart()
z60=ocp.variable('wf.w2.m_lt').getStart()

z70=ocp.variable('wf.w3.m_ga').getStart()
z80=ocp.variable('wf.w3.m_gt').getStart()
z90=ocp.variable('wf.w3.m_lt').getStart()

z100=ocp.variable('wf.w4.m_ga').getStart()
z110=ocp.variable('wf.w4.m_gt').getStart()
z120=ocp.variable('wf.w4.m_lt').getStart()

z130=ocp.variable('wf.w5.m_ga').getStart()
z140=ocp.variable('wf.w5.m_gt').getStart()
z150=ocp.variable('wf.w5.m_lt').getStart()

z160=ocp.variable('wf.f1.m_gp').getStart()
z170=ocp.variable('wf.f1.m_op').getStart()
z180=ocp.variable('wf.f1.m_wp').getStart()
z190=ocp.variable('wf.f1.m_gr').getStart()
z200=ocp.variable('wf.f1.m_or').getStart()
z210=ocp.variable('wf.f1.m_wr').getStart()

z220=ocp.variable('wf.f2.m_gp').getStart()
z230=ocp.variable('wf.f2.m_op').getStart()
z240=ocp.variable('wf.f2.m_wp').getStart()
z250=ocp.variable('wf.f2.m_gr').getStart()
z260=ocp.variable('wf.f2.m_or').getStart()
z270=ocp.variable('wf.f2.m_wr').getStart()


# BEGIN SCALING
# Get scaling factors values
def getNominal(v):
  ret = []
  for i in v:
    ret.append(i.getNominal())
  return array(ret,dtype=SX)

z_sca = getNominal(ocp.xd)

z_i=[0,z10,z20,z30,z40,z50,z60,z70,z80,z90,z100,z110,z120,z130,z140,z150,z160,z170,\
	z180,z190,z200,z210,z220,z230,z240,z250,z260,z270]
z_i=(z_i/z_sca)

# END SCALING

# Create initial guess vector
initU	=ones(nk*nWells)*1
initV2=transpose(tile([	z_i[0],z_i[1],z_i[2],\
						z_i[3],z_i[4],z_i[5],\
						z_i[6],z_i[7],z_i[8],\
						z_i[9],z_i[10],z_i[11],\
						z_i[12],z_i[13],z_i[14],\
						z_i[15],z_i[16],z_i[17],z_i[18],z_i[19],z_i[20],\
						z_i[21],z_i[22],z_i[23],z_i[24],z_i[25],z_i[26],z_i[27]],(nk)))


initVec=concatenate((initU,initV2))
VINIT=initVec



# Constraint functions
g = [];

g_min=NP.zeros([nk,nStates])
g_max=NP.zeros([nk,nStates])

# Initial condition constr.
var0 = vertcat((Z0[0],Z1[0],Z2[0],Z3[0],Z4[0],Z5[0],Z6[0],Z7[0],Z8[0],Z9[0],Z10[0],Z11[0],Z12[0],Z13[0],Z14[0],Z15[0],\
				Z16[0],Z17[0],Z18[0],Z19[0],Z20[0],Z21[0],Z22[0],Z23[0],Z24[0],Z25[0],Z26[0],Z27[0]))
g.append(var0-vertcat(z_i))
					

# Build up a graph of integrator calls
for k in range(nk):
	Zk = vertcat((Z0[k],Z1[k],Z2[k],Z3[k],Z4[k],Z5[k],Z6[k],Z7[k],Z8[k],Z9[k],Z10[k],Z11[k],Z12[k],Z13[k],Z14[k],Z15[k],\
	Z16[k],Z17[k],Z18[k],Z19[k],Z20[k],Z21[k],Z22[k],Z23[k],Z24[k],Z25[k],Z26[k],Z27[k]))

	if k!= nk-1:
		Zk_next = vertcat((Z0[k+1],Z1[k+1],Z2[k+1],Z3[k+1],Z4[k+1],Z5[k+1],Z6[k+1],Z7[k+1],Z8[k+1],Z9[k+1],Z10[k+1],Z11[k+1],Z12[k+1],Z13[k+1],Z14[k+1],Z15[k+1],\
		Z16[k+1],Z17[k+1],Z18[k+1],Z19[k+1],Z20[k+1],Z21[k+1],Z22[k+1],Z23[k+1],Z24[k+1],Z25[k+1],Z26[k+1],Z27[k+1],))


	Zk_end = f_d.call([Zk,vertcat((U[k],U[k+nk*1],U[k+nk*2],U[k+nk*3],U[k+nk*4]))])[0]	# Integrator call

	if k!= nk-1:
		g.append(Zk_next - Zk_end)		# Continuity constraints



  
# Objective function: L(T)
F = MXFunction([V],[Zk_end[0]])			# Zk_end[0] = Objective function: zc(tf)=zc(tN)


g2=U[0:1*nk]+U[1*nk:2*nk]+U[2*nk:3*nk]+U[3*nk:4*nk]+U[4*nk:5*nk]

g1=vertcat(g)							# Concatenate initial and all continuity constraints
g3=vertcat((g1,g2))						# Concatenate g1 and capacity constraints

G = MXFunction([V],[g3])				# Constraint fcn g(V)


solver = IpoptSolver(F,G)				# Provide obj. and constr. to IPOPT


solver.setOption("tol",IPOPT_TOL)				# IPOPT tolerance
solver.setOption("max_iter",IPOPT_MAX_ITER)		# IPOPT maximum iterations
solver.init()

# Set bounds and initial guess
solver.setInput(VMIN,  NLP_LBX)					# v_L	
solver.setInput(VMAX,  NLP_UBX)					# v_U
solver.setInput(VINIT, NLP_X_INIT)				# Initial guess


G_MIN = vertcat((NP.concatenate(g_min),uMinCapacity))	# g_L: [z0 0 ... 0 U_{tot,min} ... U_{tot,min}]
G_MAX = vertcat((NP.concatenate(g_max),uMaxCapacity))	# g_U: [z0 0 ... 0 U_{tot,max} ... U_{tot,max}]


solver.setInput(G_MIN,NLP_LBG)
solver.setInput(G_MAX,NLP_UBG)

solver.solve()		# Solve NLP


# Obtain results from optimization
v_opt = solver.output(NLP_X_OPT)

u_opt1 = v_opt[0:nk]
u_opt2 = v_opt[nk:nk*2]
u_opt3 = v_opt[nk*2:nk*3]
u_opt4 = v_opt[nk*3:nk*4]
u_opt5 = v_opt[nk*4:nk*5]

x0_opt = v_opt[nk*nWells+0::nStates]
x1_opt = v_opt[nk*nWells+1::nStates]
x2_opt = v_opt[nk*nWells+2::nStates]

x3_opt = v_opt[nk*nWells+3::nStates]
x4_opt = v_opt[nk*nWells+4::nStates]
x5_opt = v_opt[nk*nWells+5::nStates]

x6_opt = v_opt[nk*nWells+6::nStates]
x7_opt = v_opt[nk*nWells+7::nStates]
x8_opt = v_opt[nk*nWells+8::nStates]

x9_opt = v_opt[nk*nWells+9::nStates]
x10_opt = v_opt[nk*nWells+10::nStates]
x11_opt = v_opt[nk*nWells+11::nStates]

x12_opt = v_opt[nk*nWells+12::nStates]
x13_opt = v_opt[nk*nWells+13::nStates]
x14_opt = v_opt[nk*nWells+14::nStates]



u_opt_1=transpose(kron(transpose(NP.array(u_opt1)),ones(100)))
u_opt_2=transpose(kron(transpose(NP.array(u_opt2)),ones(100)))
u_opt_3=transpose(kron(transpose(NP.array(u_opt3)),ones(100)))
u_opt_4=transpose(kron(transpose(NP.array(u_opt4)),ones(100)))
u_opt_5=transpose(kron(transpose(NP.array(u_opt5)),ones(100)))




fmu_name = compile_fmu("MODEL.full","modell.mop")		# Compile model for simulation purpose

vdp = FMUModel(fmu_name)

t = NP.linspace(0.,tf,size(u_opt_1)) 

# Create u profile for inputting to simulation
u_traj = NP.transpose(NP.vstack((t,NP.transpose(u_opt_1),NP.transpose(u_opt_2),NP.transpose(u_opt_3),NP.transpose(u_opt_4),NP.transpose(u_opt_5))))
input_object = (['u1','u2','u3','u4','u5'], u_traj)
res = vdp.simulate(final_time = tf,input=input_object)

w_gout1 = res['f1.w_gout']
w_oout1 = res['f1.w_oout']
w_wout1 = res['f1.w_wout']

w_gout2 = res['f2.w_gout']
w_oout2 = res['f2.w_oout']
w_wout2 = res['f2.w_wout']

u1 = res['u1']
u2 = res['u2']
u3 = res['u3']
u4 = res['u4']
u5 = res['u5']


t = res['time']

clf

subplot(411)
plot(t,w_gout1,'g')
plot(t,w_oout1,'k')
plot(t,w_wout1,'b')

subplot(412)
plot(t,w_gout2,'g')
plot(t,w_oout2,'k')
plot(t,w_wout2,'b')

subplot(413)
plot(t,u1)
plot(t,u2)
plot(t,u3)
plot(t,u4)
plot(t,u5)

subplot(414)
plot(t,u1+u2+u3+u4+u5)

show()

# Export data to MATLAB
# scipy.io.savemat('C:/', \
			# dict(	t_ms=t,\
					# w_gout1_ms=w_gout1,w_oout1_ms=w_oout1,w_wout1_ms=w_wout1,\
					# w_gout2_ms=w_gout2,w_oout2_ms=w_oout2,w_wout2_ms=w_wout2,\
					# u_gi1_ms=u_opt_1,u_gi2_ms=u_opt_2,u_gi3_ms=u_opt_3,u_gi4_ms=u_opt_4,u_gi5_ms=u_opt_5))


