#
# Well-Flowline: Single shooting
# Master thesis attachment
# Konstantin Nalum 2013
#

import os
import sys
import numpy as NP
from numpy import *
import matplotlib.pyplot as plt
import zipfile
import time
import scipy.io

os.system("CLS")
from pymodelica import compile_jmu
from pyjmi import JMUModel
from casadi import *
from matplotlib.pyplot import *
from pymodelica import compile_fmu
from pyfmi import FMUModel

# Compile model and generate XML equations
curr_dir = os.path.dirname(os.path.abspath(__file__))
jmu_name = compile_jmu("MODEL.ocp", curr_dir+"/modell.mop",'optimica','ipopt',{'generate_xml_equations':True, 'generate_fmi_me_xml':False})
sfile = zipfile.ZipFile(curr_dir+'/MODEL_ocp.jmu','r')
mfile = sfile.extract('modelDescription.xml','.')

# Obtain Symbolic OCP representation
ocp = SymbolicOCP()
ocp.parseFMI('modelDescription.xml',{'scale_variables':True,'sort_equations':False})
ocp.sortType()
ocp.eliminateDependent()
ocp.makeExplicit()


tf=6*3600		# Final time
N=60			# Control input discretization
nWells=5		# Number of wells
nFlowlines=2	# Number of flowlines


# U_{tot,min} (partitoned over time)
LBG = concatenate((	1*ones(N/4),\
					1*ones(N/4),\
					1*ones(N/4),\
					1*ones(N/4),\
					))

# U_{tot,max} (partitoned over time)
UBG = concatenate((	7*ones(N/4),\
					7*ones(N/4),\
					7*ones(N/4),\
					7*ones(N/4),\
					))

u_min 		= 	ocp.variable('u_min').getStart()		# Lift-gas lower bound
u_max 		= 	ocp.variable('u_max').getStart()		# Lift-gas upper bound
CVODES_ABS_TOL = 1e-5									# CVODES absolute tolerance
CVODES_REL_TOL = 1e-5									# CVODES relative tolerance
IPOPT_TOL = 1e-1										# IPOPT tolerance: tol

# ------------------------------------


def toArray(v, der=False):	# Convert symbolic OCP variables to another format
  ret = []
  for i in v:
    if der:
      ret.append(i.der())
    else:
      ret.append(i.var())
  return array(ret,dtype=SX)
  
z = toArray(ocp.xd)				# States
u = toArray(ocp.u)				# Control inputs
f=ocp.ode						# ODE RHS (model equations)

# Generate ode: f=(z,u)
ode=SXFunction(daeIn(z,[],u,[],[]),daeOut(f))


f_d = CVodesIntegrator(ode)						# Instantiate a CVodesIntegrator
f_d.setOption("abstol",CVODES_ABS_TOL)			# Absolute tolerance
f_d.setOption("reltol",CVODES_REL_TOL)			# Relative tolerance
f_d.setOption("steps_per_checkpoint",1000)		# Steps per checkpoint
f_d.setOption("tf",tf/N)						# Integration length for each u^l
f_d.init()


U_TOT=msym("U",nWells*N)						# All control inputs in a vector

# INITIAL CONDITIONS

# BEGIN get scaling values
def getNominal(v):
  ret = []
  for i in v:
    ret.append(i.getNominal())
  return array(ret,dtype=SX)

z_sca = getNominal(ocp.xd)


z10=ocp.variable('wf.w1.m_ga').getStart()
z20=ocp.variable('wf.w1.m_gt').getStart()
z30=ocp.variable('wf.w1.m_lt').getStart()

z40=ocp.variable('wf.w2.m_ga').getStart()
z50=ocp.variable('wf.w2.m_gt').getStart()
z60=ocp.variable('wf.w2.m_lt').getStart()

z70=ocp.variable('wf.w3.m_ga').getStart()
z80=ocp.variable('wf.w3.m_gt').getStart()
z90=ocp.variable('wf.w3.m_lt').getStart()

z100=ocp.variable('wf.w4.m_ga').getStart()
z110=ocp.variable('wf.w4.m_gt').getStart()
z120=ocp.variable('wf.w4.m_lt').getStart()

z130=ocp.variable('wf.w5.m_ga').getStart()
z140=ocp.variable('wf.w5.m_gt').getStart()
z150=ocp.variable('wf.w5.m_lt').getStart()

z160=ocp.variable('wf.f1.m_gp').getStart()
z170=ocp.variable('wf.f1.m_op').getStart()
z180=ocp.variable('wf.f1.m_wp').getStart()
z190=ocp.variable('wf.f1.m_gr').getStart()
z200=ocp.variable('wf.f1.m_or').getStart()
z210=ocp.variable('wf.f1.m_wr').getStart()

z220=ocp.variable('wf.f2.m_gp').getStart()
z230=ocp.variable('wf.f2.m_op').getStart()
z240=ocp.variable('wf.f2.m_wp').getStart()
z250=ocp.variable('wf.f2.m_gr').getStart()
z260=ocp.variable('wf.f2.m_or').getStart()
z270=ocp.variable('wf.f2.m_wr').getStart()



z_i=[0,z10,z20,z30,z40,z50,z60,z70,z80,z90,z100,z110,z120,z130,z140,z150,z160,z170,\
	z180,z190,z200,z210,z220,z230,z240,z250,z260,z270]


z_i=(z_i/z_sca)	# z_i (unscaled initial values), x_sca (nominal values)


Z=MX([	z_i[0],z_i[1],z_i[2],\
		z_i[3],z_i[4],z_i[5],\
		z_i[6],z_i[7],z_i[8],\
		z_i[9],z_i[10],z_i[11],\
		z_i[12],z_i[13],z_i[14],\
		z_i[15],z_i[16],z_i[17],z_i[18],z_i[19],z_i[20],\
		z_i[21],z_i[22],z_i[23],z_i[24],z_i[25],z_i[26],z_i[27]])

ZP = MX()
# END SCALING....

# Build up a graph of integrator calls
for k in range(N):
	[Z,Xp,_,_] = f_d.call([Z,vertcat((U_TOT[k],U_TOT[k+1*N],U_TOT[k+2*N],U_TOT[k+3*N],U_TOT[k+4*N])),ZP])
	
F=MXFunction([U_TOT],[Z[0]])	# Z[0] = Objective function: zc(tf)=zc(tN)

g2=U_TOT[0:1*N]+U_TOT[1*N:2*N]+U_TOT[2*N:3*N]+U_TOT[3*N:4*N]+U_TOT[4*N:5*N]	

G=MXFunction([U_TOT],[g2])	# Constraint fcn

solver=IpoptSolver(F,G)				# Provide obj. and constr. to IPOPT
solver.setOption("tol",IPOPT_TOL)	# IPOPT tolerance
solver.setOption("max_iter",150)	# IPOPT maximum iterations
solver.init()

solver.setInput(nWells*N*[u_min],NLP_LBX)	# u_L
solver.setInput(nWells*N*[u_max],NLP_UBX)	# u_U

u_guess = 1.1
solver.setInput(nWells*N*[u_guess],NLP_X_INIT)	# Initial guess



solver.setInput(LBG,NLP_LBG)	# g_L
solver.setInput(UBG,NLP_UBG)	# g_U

solver.solve()	# Solve NLP


# Obtain results from optimization
u_opt = NP.array(solver.output(NLP_X_OPT))

u_opt1=u_opt[0*N:1*N]
u_opt2=u_opt[1*N:2*N]
u_opt3=u_opt[2*N:3*N]
u_opt4=u_opt[3*N:4*N]
u_opt5=u_opt[4*N:5*N]


# Compile model for simulation purpose
fmu_name = compile_fmu("MODEL.full","modell.mop")

wfmodel = FMUModel(fmu_name)

u_opt_1=transpose(kron(transpose(NP.array(u_opt1)),ones(100)))
u_opt_2=transpose(kron(transpose(NP.array(u_opt2)),ones(100)))
u_opt_3=transpose(kron(transpose(NP.array(u_opt3)),ones(100)))
u_opt_4=transpose(kron(transpose(NP.array(u_opt4)),ones(100)))
u_opt_5=transpose(kron(transpose(NP.array(u_opt5)),ones(100)))

t = NP.linspace(0,tf,size(u_opt_1)) 

# Create u profile for inputting to simulation
u_traj = NP.transpose(NP.vstack((t,NP.transpose(u_opt_1),NP.transpose(u_opt_2),NP.transpose(u_opt_3),NP.transpose(u_opt_4),NP.transpose(u_opt_5))))
input_object = (['u1','u2','u3','u4','u5'], u_traj)
res = wfmodel.simulate(final_time = tf,input=input_object)

w_gout1 = res['f1.w_gout']
w_oout1 = res['f1.w_oout']
w_wout1 = res['f1.w_wout']

w_gout2 = res['f2.w_gout']
w_oout2 = res['f2.w_oout']
w_wout2 = res['f2.w_wout']

u1 = res['u1']
u2 = res['u2']
u3 = res['u3']
u4 = res['u4']
u5 = res['u5']


t = res['time']

# Plot optimized and simulated profiles
clf

subplot(411)
plot(t,w_gout1,'g')
plot(t,w_oout1,'k')
plot(t,w_wout1,'b')

subplot(412)
plot(t,w_gout2,'g')
plot(t,w_oout2,'k')
plot(t,w_wout2,'b')

subplot(413)
step(t,u1)
step(t,u2)
step(t,u3)
step(t,u4)
step(t,u5)

subplot(414)
step(t,u1+u2+u3+u4+u5)


show()

# Export data to MATLAB

# scipy.io.savemat('C:/', \
			# dict(	t_ss=t,\
					# w_gout1_ss=w_gout1,w_oout1_ss=w_oout1,w_wout1_ss=w_wout1,\
					# w_gout2_ss=w_gout2,w_oout2_ss=w_oout2,w_wout2_ss=w_wout2,\
					# u_gi1_ss=u_opt1,u_gi2_ss=u_opt2,u_gi3_ss=u_opt3,u_gi4_ss=u_opt4,u_gi5_ss=u_opt5))

